// gml_Errors.cpp --
//
//    See "gmlErrors.h".
//
//  Copyright (c) 1995-2002 CLIPS-IMAG
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on July, 1995 (FB).

#include "gml_Errors.h"

#include "gml_Alloc.h"
#include "gml_DynamicArray.h"

static char   cErrorIncorrectCode[]         = "incorrect error code";
static char   cErrorNotInitialized[]        = "error message invalid (not initialized, or no more memory)";

gml_TError    gml_gErrorMemFull             = (gml_TError)-1;
gml_TError    gml_gErrorWrongParameterVal   = (gml_TError)-1;
gml_TError    gml_gDivideByZero             = (gml_TError)-1;
gml_TError    gml_gNotInitialized           = (gml_TError)-1;
gml_TError    gml_gErrorFileOpenRead        = (gml_TError)-1;
gml_TError    gml_gErrorFileOpenWrite       = (gml_TError)-1;
gml_TError    gml_gErrorFileRead            = (gml_TError)-1;
gml_TError    gml_gErrorFileWrite           = (gml_TError)-1;


typedef gml_TDynamicArray<char*>        gml_TStringDynArray;

static gml_TStringDynArray* gArray      = (gml_TStringDynArray*)NULL;
static int                  gNbErrors   = 0;



// gml_ErrorInit --

int gml_ErrorInit ()
{
  gml_TError      err;
  char**          message;
  gml_TBlockSize  index;
  
  err = gml_New (gArray);
  if (err != 0)
    return 1;
  
  err = gArray->Init (20, 20);
  if (err != 0) {
    gml_Free (gArray);
    return 1;
  }

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  if (index != (gml_TBlockSize)gml_cNoError)
    return 1;
  *message = "no error";
  gNbErrors++;

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  gml_gErrorMemFull = (gml_TError)index;
  *message = "no more memory";
  gNbErrors++;

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  gml_gErrorWrongParameterVal = (gml_TError)index;
  *message = "incorrect parameter value";
  gNbErrors++;

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  gml_gDivideByZero = (gml_TError)index;
  *message = "division by 0 or singularity";
  gNbErrors++;

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  gml_gNotInitialized = (gml_TError)index;
  *message = "attempt to use an uninitilialized object";
  gNbErrors++;

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  gml_gErrorFileOpenRead = (gml_TError)index;
  *message = "could not open file for reading";
  gNbErrors++;

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  gml_gErrorFileOpenWrite = (gml_TError)index;
  *message = "could not open file for writing";
  gNbErrors++;

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  gml_gErrorFileRead = (gml_TError)index;
  *message = "error while reading from file";
  gNbErrors++;

  if ((message = gArray->GetElem (&index)) == (char**)NULL)
    return 1;
  gml_gErrorFileWrite = (gml_TError)index;
  *message = "error while writing to file";
  gNbErrors++;

  return 0;
}



// gml_ErrorDispose --

extern "C"
void gml_ErrorDispose ()
{
  gml_TestDisposeAndDelete (gArray);
  gNbErrors = 0;
}



// gml_ErrorNew --

extern "C"
gml_TError gml_ErrorNew (char* message)
{
  int             res;
  char**          messagePtr;
  gml_TBlockSize  index;

  if (gArray == (gml_TStringDynArray*)NULL) {
    res = gml_ErrorInit ();
    if (res != 0)
      return (gml_TError)-1;
  }

  messagePtr = gArray->GetElem (&index);
  if (messagePtr == (char**)NULL)
    return (gml_TError)-1;

  *messagePtr = message;
  gNbErrors ++;
  
  return (gml_TError)index;
}



// gml_ErrorGetMessage --

extern "C"
char* gml_ErrorGetMessage (gml_TError error)
{
  int res;

  if (gArray == (gml_TStringDynArray*)NULL) {
    res = gml_ErrorInit ();
    if (res != 0)
      return cErrorNotInitialized;
  }

  if (error == -1)
    return cErrorNotInitialized;
  
  if ((error < 0) || (error >= gNbErrors))
    return cErrorIncorrectCode;

  return (gArray->Base())[error];
}


