// gmlTcl_Class.h --
//
//    Declares the gmlTcl_TClass class. An object of this class represents a
//    Tcl-level gmlObject class.
//
//  Copyright (c) 2007 LIG/IIHM
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on June 21, 2007 (FB).

#ifndef __GMLTCL_CLASS__
#define __GMLTCL_CLASS__

extern "C" {
 #include "tcl.h"
}

#include "gml_List.h"
#include "gmlTcl_Entity.h"

class gmlTcl_TObject;
class gmlTcl_TMethod;

class gmlTcl_TClass : public gmlTcl_TEntity
{
  public:

    // Public object methods

    int                     Init                   (Tcl_Interp*         interp,
                                                    Tcl_Obj*            name);
    void                    Dispose                ();

    gml_TError              CreateRecursiveAncestorList
                                                   (gml_TList**         list);

    int                     IsAncestor             (gmlTcl_TClass*      ancestor);

    int                     Inherit                (Tcl_Interp*         interp,
                                                    gmlTcl_TClass*      superClass);

    void                    DeInherit              (gmlTcl_TClass*      superClass);

    int                     CommandHandler         (Tcl_Interp*         interp,
                                                    int                 objc,
                                                    Tcl_Obj*            objv[]);

    void                    RegisterObject         (gmlTcl_TObject*     theObject);
    void                    DeregisterObject       (gmlTcl_TObject*     theObject);

    // flags & 1 -> method is the constructor
    // flags & 2 -> method is the destructor
    void                    RegisterMethod         (gmlTcl_TMethod*     theMethod,
                                                    int                 flags);
    void                    DeregisterMethod       (gmlTcl_TMethod*     theMethod,
                                                    int                 flags);

    // flags & 1 -> search in this class methods
    // flags & 2 -> search in this class ancestor methods

    gmlTcl_TMethod*         FindMethod             (Tcl_Obj*            methodNameObj,
                                                    Tcl_Interp*         interp            = (Tcl_Interp*)NULL,
                                                    int                 flags             = 3);

    gml_TError              FullMethodList         (gml_TList**         list,
                                                    int                 superPrefix);


    // Accessors, Modifiers

    gml_TList*              ObjectList ()          { return fObjectList;      }
    Tcl_HashTable*          MethodTable ()         { return &fMethodTable;    }

    gmlTcl_TMethod*         Constructor ()         { return fConstructor;     }
    gmlTcl_TMethod*         Destructor ()          { return fDestructor;      }

    gml_TList*              Ancestors ()           { return fAncestors;       }
    gml_TList*              Specializations ()     { return fSpecializations; }


  protected:

    gml_TError              AddMethodsToList       (gml_TList*          alreadyIn,
                                                    gml_TList*          result,
                                                    int                 prefix,
                                                    int                 recursive);

    // Object attributes

    gml_TList*              fObjectList;      // the list of instances (objects) of this class.
    Tcl_HashTable           fMethodTable;     // the hash table containing all methods of this class,
                                              //    from string name to method pointer.
    int                     fMethodTableInit; // true if "fMethodTable" has been initialized.
    gmlTcl_TMethod*         fConstructor;     // pointer to constructor method if there is one, or NULL.
    gmlTcl_TMethod*         fDestructor;      // pointer to destructor method if there is one, or NULL.

    gml_TList*              fAncestors;       // list of ancestor classes.
    gml_TList*              fSpecializations; // list of classes specializing this class.
};


#endif
