// gmlTcl_Obj_Init.cpp --
//
//    C implementation of the gml object constructs in Tcl scripts.
//
//  Copyright (c) 2007 LIG/IIHM
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on June 16, 2007 (FB).


extern "C" {
 #include "tcl.h"
 #include "string.h"
 #include "stdlib.h"
 #include "stdio.h"
}

#include "gml_Alloc.h"

#include "gmlTcl_Obj_Init.h"

#include "gmlTcl_Obj_Internal.h"
#include "gmlTcl_Entity.h"
#include "gmlTcl_Object.h"
#include "gmlTcl_Method.h"
#include "gmlTcl_Class.h"
#include "gmlTcl_gmlObject.h"

char*                   gmlTcl_kEntityTypeStrings[]   = { "class", "object" };

Tcl_Interp*             gmlTcl_gObj_Interp            = (Tcl_Interp*)NULL;
Tcl_HashTable           gmlTcl_gObj_ClassTable;
Tcl_HashTable           gmlTcl_gObj_ObjectTable;

static Tcl_Command      gmlTcl_gObj_MethodCmd         = (Tcl_Command)NULL;
static Tcl_Command      gmlTcl_gObj_gmlObjectCmd      = (Tcl_Command)NULL;
static Tcl_Command      gmlTcl_gObj_ThisCmd           = (Tcl_Command)NULL;
static Tcl_Command      gmlTcl_gObj_InheritCmd        = (Tcl_Command)NULL;

static char             gmlTcl_gObj_RootClassScript[] = "method gmlObjRootClass dispose {} {\n  gmlObject delete object $objName\n}\nmethod gmlObjRootClass attribute { name } {\n  return $this($name)\n}\nmethod gmlObjRootClass setAttribute { name value } {\n  return [set this($name) $value]\n}\nmethod gmlObjRootClass attributes { args } {\n  set res [list]\n  foreach attribute $args {\n    lappend res $this($attribute)\n  }\n  return $res\n}\nmethod gmlObjRootClass newName { name } {\n  gmlObjNewObject $objName [gmlObject info classofobject $objName] $name\n}\nmethod gmlObjRootClass unknown { method args } {\nglobal gmlObject\n\n  error \"invalid method name \\\"$method\\\": should be one of \\\"[gmlObjBuildMethodList [gmlObject info classofobject $objName]]\\\"\"\n}";
gmlTcl_TClass*          gmlTcl_gObj_RootClass         = (gmlTcl_TClass*)NULL;

static char             gmlTcl_gObj_VarExistProcName[]        = "gmlGlobalVarExists";
static char             gmlTcl_gObj_VarExistProcDefinition[]  = "proc gmlGlobalVarExists { varName } { global $varName ; return [info exists $varName] }\n";

static char             gmlTcl_gObj_CloneArrayProcName[]        = "gmlCloneGlobalArray";
static char             gmlTcl_gObj_CloneArrayProcDefinition[]  = "proc gmlCloneGlobalArray { src dst } { global $src $dst ; array set $dst [array get $src] }\n";



// gmlTcl_Obj_Interp --

Tcl_Interp* gmlTcl_Obj_Interp ()
{
  return gmlTcl_gObj_Interp;
}



// UpdateStringNames --

int gmlTcl_Obj_UpdateStringNames (char*               name,
                                  char**              shortBuffer,
                                  char**              protectedBuffer,
                                  Tcl_Interp*         interp)
{
  char*               stringName          = (char*)NULL;
  char*               protectedString     = (char*)NULL;
  int                 nameLen             = strlen (name);
  Tcl_Obj*            listObj;
  Tcl_Obj*            listArray[1];
  char*               listStr;
  int                 listStrLen;

  // Create the string name.
  
  stringName          = (char*)malloc (nameLen + 1);
  if (stringName == (char*)NULL)
    goto error;

  strcpy (stringName, name);

  // Create the protected string name: the string name enclosed in {} if it contains
  //  special characters such as space.
  
  listArray[0]  = Tcl_NewStringObj (name, -1);
  listObj       = Tcl_NewListObj (1, listArray);
  listStr       = Tcl_GetStringFromObj(listObj, &listStrLen);

  protectedString     = (char*)malloc (listStrLen + 1);
  if (protectedString == (char*)NULL)
    goto error;

  strcpy (protectedString, listStr);
  Tcl_DecrRefCount (listObj);

  gml_TestAndFree (*shortBuffer);
  gml_TestAndFree (*protectedBuffer);

  *shortBuffer        = stringName;
  *protectedBuffer    = protectedString;

  return TCL_OK;

error:

  if (interp != (Tcl_Interp*)NULL) {
    Tcl_AppendResult (interp, "could not create string names for \"",
                              name, "\": not enough memory", (char*)NULL);
  }
  gml_TestAndFree (stringName);
  gml_TestAndFree (protectedString);

  return TCL_ERROR;
}




// gmlTcl_Obj_AppendOptionList --

void gmlTcl_Obj_AppendOptionList (Tcl_Interp*         interp,
                                  gml_TList*          methodList)
{
  gml_TListElem*        elem    = methodList->Head ();
  gml_TListElem*        next;

  if (elem == (gml_TListElem*)NULL)
    return;

  if (elem->Next () == (gml_TListElem*)NULL) {

    // There is only one element in the list.

    Tcl_AppendResult (interp, (char*)(elem->Value ()), (char*)NULL);
    return;
  }

  // There are at least two elements in the list.

  while (elem != (gml_TListElem*)NULL) {
    next = elem->Next ();
    if (next == (gml_TListElem*)NULL)
      Tcl_AppendResult (interp, "or ", (char*)(elem->Value ()), (char*)NULL);
    else      
      Tcl_AppendResult (interp, (char*)(elem->Value ()), ", ", (char*)NULL);

    elem = next;
  }
}



// gmlTcl_Obj_EntityListToStringNameObjList --

int gmlTcl_Obj_EntityListToStringNameObjList (Tcl_Interp*         interp,
                                              gml_TList*          list,
                                              Tcl_Obj**           resObj,
                                              int                 sorted)
{
  int                 res           = TCL_OK;
  gml_TError          err           = gml_cNoError;
  Tcl_Obj*            theRes        = (Tcl_Obj*)NULL;
  gml_TList*          processed     = (gml_TList*)NULL;
  gml_TListElem*      elem;

  if (!sorted)
    processed = list;

  else {

    // Create a new list in order to sort it.

    err = gml_NewAndInit (processed);
    if (err != gml_cNoError) {
      Tcl_AppendResult (interp, "could not create list of names: no more memory", (char*)NULL);
      return TCL_ERROR;
    }

    elem = list->Head ();
    while ((err == gml_cNoError) && (elem != (gml_TListElem*)NULL)) {
      err = processed->AppendValue ((void*)(((gmlTcl_TEntity*)(elem->Value ()))->StringName ()));
      if (err == gml_cNoError)
        elem = elem->Next ();
    }
    if (err == gml_cNoError)
      err = processed->Sort ();
  }

  if (err == gml_cNoError) {
    theRes = Tcl_NewListObj (0, (Tcl_Obj**)NULL);
    elem = processed->Head ();
    while ((res == TCL_OK) && (elem != (gml_TListElem*)NULL)) {
      if (sorted)
        res = Tcl_ListObjAppendElement (interp, theRes, Tcl_NewStringObj ((char*)(elem->Value ()), -1));
      else
        res = Tcl_ListObjAppendElement (interp, theRes, Tcl_NewStringObj ((((gmlTcl_TEntity*)(elem->Value ()))->StringName ()), -1));

      if (res == TCL_OK)
        elem = elem->Next ();
    }
    if (res == TCL_OK)
      *resObj = theRes;
    else {
      Tcl_DecrRefCount (theRes);
      Tcl_AppendResult (interp, "could not create name list: error while adding elements to the list",
                                (char*)NULL);
    }
  }

  if (sorted)
    gml_TestDisposeAndDelete (processed);

  if (err != gml_cNoError) {
    Tcl_AppendResult (interp, "could not create list of names: ", gml_ErrorGetMessage (err), (char*)NULL);
    res = TCL_ERROR;
  }

  return res;
}

// gmlTcl_Obj_HashToList --

int gmlTcl_Obj_HashToList  (Tcl_Interp*         interp,
                            Tcl_HashTable*      table,
                            gml_TList**         list,
                            gml_TListSortFunc*  sortFunc)
{
  gml_TError          err           = gml_cNoError;
  gml_TList*          theList       = (gml_TList*)NULL;
  Tcl_HashEntry*      entry;
  Tcl_HashSearch      search;

  // Create a list from the hash table elements

  err = gml_NewAndInit (theList);
  if (err != gml_cNoError) {
    Tcl_AppendResult (interp, "could not create list from hash table: no more memory", (char*)NULL);
    goto error;
  }

  entry = Tcl_FirstHashEntry (table, &search);
  while ((err == gml_cNoError) && (entry != (Tcl_HashEntry*)NULL)) {
    err = theList->AppendValue ((void*)Tcl_GetHashKey (table, entry));
    if (err == gml_cNoError)
      entry = Tcl_NextHashEntry (&search);
  }
  if (err != gml_cNoError) {
    Tcl_AppendResult (interp, "could not create list from hash table: no more memory", (char*)NULL);
    goto error;
  }
  
  // Sort the list if required

  if (sortFunc != (gml_TListSortFunc*)NULL)
    err = theList->Sort (sortFunc);

  if (err != gml_cNoError) {
    Tcl_AppendResult (interp, "could not create list from hash table: error while sorting the list", (char*)NULL);
    goto error;
  }

  *list = theList;

  return TCL_OK;

error:

  gml_TestDisposeAndDelete (theList);
  return TCL_ERROR;
}

// gmlTcl_Obj_ListToObj --

int gmlTcl_Obj_ListToObj (Tcl_Interp*         interp,
                          gml_TList*          list,
                          Tcl_Obj**           resObj)
{
  int                 res           = TCL_OK;
  Tcl_Obj*            theRes;
  gml_TListElem*      elem          = list->Head ();

  theRes = Tcl_NewListObj (0, (Tcl_Obj**)NULL);

  while ((res == TCL_OK) && (elem != (gml_TListElem*)NULL)) {
    res = Tcl_ListObjAppendElement (interp, theRes, Tcl_NewStringObj ((char*)(elem->Value ()), -1));
    if (res == TCL_OK)
      elem = elem->Next ();
  }
  if (res != TCL_OK) {
    Tcl_DecrRefCount (theRes);
    Tcl_AppendResult (interp, "could not create list: error while adding elements to the list",
                              (char*)NULL);
  } else
    *resObj = theRes;

  return res;
}

// gmlTcl_Obj_StringHashTableToObj --

int gmlTcl_Obj_StringHashTableToObj  (Tcl_Interp*         interp,
                                      Tcl_HashTable*      table,
                                      Tcl_Obj**           resObj,
                                      int                 sorted)
{
  int                 res           = TCL_OK;
  gml_TList*          list          = (gml_TList*)NULL;
  Tcl_HashEntry*      entry;
  Tcl_HashSearch      search;
  Tcl_Obj*            theRes;

  if (sorted) {

    // Convert hash table to a list so that it can be sorted,
    //  sort the list and create the objList result.
  
    res = gmlTcl_Obj_HashToList (interp, table, &list, gml_ListSortStringFunc);
    if (res == TCL_OK)
      res = gmlTcl_Obj_ListToObj (interp, list, resObj);

    // Free the list.

    gml_TestDisposeAndDelete (list);

    return res;
  }

  // Create the result Tcl list directly from the (unseorted) hash table keys.

  theRes = Tcl_NewListObj (0, (Tcl_Obj**)NULL);

  entry = Tcl_FirstHashEntry (table, &search);
  while ((res == TCL_OK) && (entry != (Tcl_HashEntry*)NULL)) {
    res = Tcl_ListObjAppendElement (interp, theRes, Tcl_NewStringObj (Tcl_GetHashKey (table, entry), -1));
    if (res == TCL_OK)
      entry = Tcl_NextHashEntry (&search);
  }
  if (res != TCL_OK) {
    Tcl_DecrRefCount (theRes);
    Tcl_AppendResult (interp, "could not create list from hash table: error while adding elements to the list",
                              (char*)NULL);
  } else
    *resObj = theRes;

  return res;
}



// gmlTclObjEntityFromTclObj --

inline
void* gmlTclObjEntityFromTclObj (Tcl_Obj* obj, Tcl_HashTable* table, Tcl_Interp* interp, char* entityTypeName)
{
  Tcl_HashEntry*      entry;
  char*               entityName  = Tcl_GetString (obj);

  entry = Tcl_FindHashEntry (table, entityName);
  if (entry == (Tcl_HashEntry*)NULL) {
    if (interp != (Tcl_Interp*)NULL)
      Tcl_AppendResult (interp, "there is no ", entityTypeName, " named \"",
                                entityName, "\"", (char*)NULL);
    return (void*)NULL;
  }
  else
    return (void*)Tcl_GetHashValue (entry);
}

// gmlTcl_Obj_EntityFromTclObj --

gmlTcl_TEntity* gmlTcl_Obj_EntityFromTclObj  (gmlTcl_TEntityType  type,
                                              Tcl_Obj*            obj,
                                              Tcl_Interp*         interp)
{
  if (type == gmlTcl_kClassEntityType)
    return (gmlTcl_TEntity*)gmlTclObjEntityFromTclObj (obj, &gmlTcl_gObj_ClassTable, interp, "class");
  else
    return (gmlTcl_TEntity*)gmlTclObjEntityFromTclObj (obj, &gmlTcl_gObj_ObjectTable, interp, "object");
}

// gmlTcl_Obj_ClassFromTclObj --

gmlTcl_TClass* gmlTcl_Obj_ClassFromTclObj  (Tcl_Obj* obj, Tcl_Interp* interp)
{
  return (gmlTcl_TClass*)gmlTclObjEntityFromTclObj (obj, &gmlTcl_gObj_ClassTable, interp, "class");
}

// gmlTcl_Obj_ObjectFromTclObj --

gmlTcl_TObject* gmlTcl_Obj_ObjectFromTclObj  (Tcl_Obj* obj, Tcl_Interp* interp)
{
  return (gmlTcl_TObject*)gmlTclObjEntityFromTclObj (obj, &gmlTcl_gObj_ObjectTable, interp, "object");
}




// gmlTcl_Obj_EntityHashTable --

Tcl_HashTable* gmlTcl_Obj_EntityHashTable (gmlTcl_TEntity* theEntity)
{
  if (theEntity->Type () == gmlTcl_kClassEntityType)
    return &gmlTcl_gObj_ClassTable;
  else
    return &gmlTcl_gObj_ObjectTable;
}



// gmlTcl_Obj_RegisterEntity --

void gmlTcl_Obj_RegisterEntity (gmlTcl_TEntity* theEntity)
{
  char*             name        = theEntity->StringName ();
  char*             typeString  = theEntity->TypeString ();
  Tcl_HashEntry*    entry;
  int               newEntry;

  entry = Tcl_CreateHashEntry (theEntity->Table (), name, &newEntry);

  if (!newEntry) {
    fprintf (stderr, "PANIC: can't register %s \"%s\": %s with this name is already registered\n",
                      typeString, name, typeString);
    return;
  }

  Tcl_SetHashValue (entry, (ClientData)theEntity);
}

// gmlTcl_Obj_DeregisterEntity --

void gmlTcl_Obj_DeregisterEntity (gmlTcl_TEntity* theEntity)
{
  char*             name        = theEntity->StringName ();
  char*             typeString  = theEntity->TypeString ();
  Tcl_HashEntry*    entry;

  entry = Tcl_FindHashEntry (theEntity->Table (), name);
  if (entry == (Tcl_HashEntry*)NULL) {
    fprintf (stderr, "PANIC: can't deregister %s \"%s\": this %s is not registered\n",
                      typeString, name, typeString);
    return;
  }

  Tcl_DeleteHashEntry (entry);
}




// gmlTcl_Obj_GlobalExists --

int gmlTcl_Obj_GlobalExists (Tcl_Interp* interp, char* varName, int* exists)
{
  int       res               = TCL_OK;
  int       procNameLen       = strlen (gmlTcl_gObj_VarExistProcName);
  int       varNameLen        = strlen (varName);
  char*     script;

  script = (char*)malloc (procNameLen + 7 + varNameLen + 2);
  if (script == (char*)NULL) {
    Tcl_AppendResult (interp, "could not test existance of global variable \"", varName,
                              "\": no more memory", (char*)NULL);
    return TCL_ERROR;
  }
  sprintf (script, "%s [list %s]", gmlTcl_gObj_VarExistProcName, varName);

  res = Tcl_EvalEx (interp, script, -1, TCL_EVAL_GLOBAL);
  free ((void*)script);
  if (res != TCL_OK)
    return res;

  res = Tcl_GetIntFromObj (interp, Tcl_GetObjResult (interp), exists);

  Tcl_ResetResult (interp);

  return res;
}




// gmlTcl_Obj_CloneGlobalArray --

int gmlTcl_Obj_CloneGlobalArray (Tcl_Interp* interp, char* src, char* dst)
{
  int       res               = TCL_OK;
  int       procNameLen       = strlen (gmlTcl_gObj_VarExistProcName);
  int       srcNameLen        = strlen (src);
  int       dstNameLen        = strlen (dst);
  char*     script;

  script = (char*)malloc (procNameLen + 7 + srcNameLen + 8 + dstNameLen + 2);
  if (script == (char*)NULL) {
    Tcl_AppendResult (interp, "could not clone global array \"", src,
                              "\" to \"", dst, "\": no more memory", (char*)NULL);
    return TCL_ERROR;
  }
  sprintf (script, "%s [list %s] [list %s]", gmlTcl_gObj_CloneArrayProcName, src, dst);

  res = Tcl_EvalEx (interp, script, -1, TCL_EVAL_GLOBAL);
  free ((void*)script);
  if (res != TCL_OK)
    return res;

  Tcl_ResetResult (interp);

  return res;
}




// gml_InheritObjCmd --
//
//  Called by the interpreter as the handler of the "inherit" command in a script.

extern "C"
int gml_InheritObjCmd  (ClientData    /* dummy */,
                        Tcl_Interp*   interp,
                        int           objc,
                        Tcl_Obj*      objv[])
{
  int                 res             = TCL_OK;
  int                 classCreated    = 0;
  gmlTcl_TClass*      theClass        = (gmlTcl_TClass*)NULL;
  gmlTcl_TClass*      theSuperClass;
  char*               theSuperClassStrName;
  int                 theSuperClassStrNameLen;
  char*               scriptBuffer;

  // Check arguments.

  if (objc != 3) {
    Tcl_WrongNumArgs (interp, 1, objv, "<className> <superClassName>");
    return TCL_ERROR;
  }

  theClass      = gmlTcl_Obj_ClassFromTclObj (objv[1]);
  theSuperClass = gmlTcl_Obj_ClassFromTclObj (objv[2]);

  if (theSuperClass == (gmlTcl_TClass*)NULL) {
    // The super class doesn't exist, try to auto_load it

    theSuperClassStrName  = Tcl_GetStringFromObj(objv[2], &theSuperClassStrNameLen);
    scriptBuffer  = (char*)malloc (11 + theSuperClassStrNameLen);
    if (scriptBuffer == (char*)NULL) {
      Tcl_AppendResult (interp, "no more memory", (char*)NULL);
      return TCL_ERROR;
    }
    sprintf (scriptBuffer, "auto_load %s", theSuperClassStrName);
    res = Tcl_Eval (interp, scriptBuffer);
    free ((void*)scriptBuffer);

    if (atoi (Tcl_GetStringResult (interp)) != 1) {
      Tcl_ResetResult (interp);
      Tcl_AppendResult (interp, "class \"", theSuperClassStrName,
                                "\" does not exist", (char*)NULL);
      return TCL_ERROR;
    }
    Tcl_ResetResult (interp);
    theSuperClass = gmlTcl_Obj_ClassFromTclObj (objv[2], interp);
    if (theSuperClass == (gmlTcl_TClass*)NULL)
      return TCL_ERROR;
  }

  if (theClass == (gmlTcl_TClass*)NULL) {

    // The specialization does not exist. Create it.

    theClass = new gmlTcl_TClass;
    if (theClass == (gmlTcl_TClass*)NULL) {
      Tcl_AppendResult (interp, "could not create class \"", Tcl_GetString (objv[1]),
                                "\": not enough memory", (char*)NULL);
      return TCL_ERROR;
    }
    classCreated = 1;
    res = theClass->Init (interp, objv[1]);
    if (res != TCL_OK)
      goto error;

  }

  // The specialization class already exists.
  //    Return immediately if the superclass is already one of its ancestors.
  //    Generate an error if the specialization is already an ancestor of the superclass.

  if (theClass->IsAncestor (theSuperClass))
    return TCL_OK;
  if (theSuperClass->IsAncestor (theClass)) {
    Tcl_AppendResult (interp, "class \"", theClass->StringName (),
                              "\" is ancestor of class \"", theSuperClass->StringName (),
                              "\"", (char*)NULL);
    return TCL_ERROR;
  }

  res = theClass->Inherit (interp, theSuperClass);
  if (res != TCL_OK)
    goto error;

  return res;
  
error:

  if (classCreated) {
    theClass->Dispose ();
    delete theClass;
  }

  return res;
}




// gml_ThisObjCmd --
//
//  Called by the interpreter as the handler of the "this" command in a script.

extern "C"
int gml_ThisObjCmd (ClientData    /* dummy */,
                    Tcl_Interp*   interp,
                    int           objc,
                    Tcl_Obj*      objv[])
{
  Tcl_Obj*            objName;
  gmlTcl_TObject*     theObject;

  // Find the variable "objName" in the context of the call to "this".

  objName = Tcl_GetVar2Ex (interp, "objName", (char*)NULL, 0);
  if (objName == (Tcl_Obj*)NULL) {
    Tcl_AppendResult (interp, "\"", Tcl_GetString (objv[0]),
                              "\" must be called in the context of a method", (char*)NULL);
    return TCL_ERROR;
  }

  // Find the object which name is stored in the "objName" variable.

  theObject = gmlTcl_Obj_ObjectFromTclObj (objName, interp);
  if (theObject == (gmlTcl_TObject*)NULL)
    return TCL_ERROR;

  // Make the method call.
  
  return theObject->MethodDispatchCmd (interp, objName, objc, objv);
}




// gml_MethodObjCmd --
//
//  Called by the interpreter as the handler of the "method" command in a script.

extern "C"
int gml_MethodObjCmd (ClientData    /* dummy */,
                      Tcl_Interp*   interp,
                      int           objc,
                      Tcl_Obj*      objv[])
{
  int               res           = TCL_OK;
  int               classCreated  = 0;
  gmlTcl_TClass*    theClass      = (gmlTcl_TClass*)NULL;
  gmlTcl_TMethod*   theMethod     = (gmlTcl_TMethod*)NULL;

  // Check arguments */

  if (objc != 5) {
      Tcl_WrongNumArgs(interp, 1, objv, "className methodName args body");
      return TCL_ERROR;
  }
  
  // Check if a class with that name already exists. If not, create it
  
  theClass = gmlTcl_Obj_ClassFromTclObj (objv[1]);
  
  if (theClass == (gmlTcl_TClass*)NULL) {
    theClass = new gmlTcl_TClass;
    if (theClass == (gmlTcl_TClass*)NULL) {
      Tcl_AppendResult (interp, "could not create class \"", Tcl_GetString (objv[1]),
                                "\": not enough memory", (char*)NULL);
      return TCL_ERROR;
    }
    res = theClass->Init (interp, objv[1]);
    classCreated = 1;

    if (res != TCL_OK)
      goto error;
  }

  // Create the method.

  theMethod = new gmlTcl_TMethod;
  if (theMethod == (gmlTcl_TMethod*)NULL) {
    Tcl_AppendResult (interp, "could not create method \"",
                              Tcl_GetString (objv[1]), "\": not enough memory",
                              (char*)NULL);
    goto error;
  }
  res = theMethod->Init (interp, theClass, objc, objv);
  if (res != TCL_OK)
    goto error;

  return res;

error:

  if (classCreated) {
    theClass->Dispose ();
    delete theClass;
  }

  if (theMethod != (gmlTcl_TMethod*)NULL) {
    theMethod->Dispose ();
    delete theMethod;
  }

  return res;
}




// gmlTclCleanupCommand --

static
void gmlTclCleanupCommand (ClientData clientData)
{
  Tcl_Command*      cmdPtr = (Tcl_Command*)clientData;

  *cmdPtr = (Tcl_Command)NULL;
}



// gmlTcl_Obj_Init --

extern "C"
int gmlTcl_Obj_Init (Tcl_Interp* interp)
{
  int     res = TCL_OK;

  gmlTcl_gObj_Interp = interp;
  
  Tcl_InitHashTable (&gmlTcl_gObj_ClassTable, TCL_STRING_KEYS);
  Tcl_InitHashTable (&gmlTcl_gObj_ObjectTable, TCL_STRING_KEYS);

  gmlTcl_gObj_MethodCmd       = Tcl_CreateObjCommand (interp, "method",
                                                      &gml_MethodObjCmd,
                                                      (ClientData)&gmlTcl_gObj_MethodCmd,
                                                      &gmlTclCleanupCommand);

  if (res == TCL_OK)
    gmlTcl_gObj_gmlObjectCmd  = Tcl_CreateObjCommand (interp, "gmlObject",
                                                      &gml_gmlObjectObjCmd,
                                                      (ClientData)&gmlTcl_gObj_gmlObjectCmd,
                                                      &gmlTclCleanupCommand);

  if (res == TCL_OK)
    gmlTcl_gObj_ThisCmd       = Tcl_CreateObjCommand (interp, "this",
                                                      &gml_ThisObjCmd,
                                                      (ClientData)&gmlTcl_gObj_ThisCmd,
                                                      &gmlTclCleanupCommand);

  if (res == TCL_OK)
    gmlTcl_gObj_InheritCmd    = Tcl_CreateObjCommand (interp, "inherit",
                                                      &gml_InheritObjCmd,
                                                      (ClientData)&gmlTcl_gObj_InheritCmd,
                                                      &gmlTclCleanupCommand);

  if (res == TCL_OK)
    res = Tcl_EvalEx (interp, gmlTcl_gObj_VarExistProcDefinition, -1, TCL_EVAL_GLOBAL);

  if (res == TCL_OK)
    res = Tcl_EvalEx (interp, gmlTcl_gObj_CloneArrayProcDefinition, -1, TCL_EVAL_GLOBAL);

  if (res == TCL_OK)
    res = Tcl_EvalEx (interp, gmlTcl_gObj_RootClassScript, -1, TCL_EVAL_GLOBAL);

  return res;
}

// gmlTcl_Obj_Dispose --

extern "C"
void gmlTcl_Obj_Dispose (Tcl_Interp* interp)
{
  Tcl_HashEntry*    entry       = (Tcl_HashEntry*)NULL;
  Tcl_HashSearch    search;
  gmlTcl_TClass*    theClass;
  gmlTcl_TObject*   theObject;

  // Delete the "gmlGlobalVarExists" proc.

  Tcl_DeleteCommand (interp, gmlTcl_gObj_VarExistProcName);
  Tcl_DeleteCommand (interp, gmlTcl_gObj_CloneArrayProcName);

  // Deregister the gmlObject tcl commands.

  if (gmlTcl_gObj_InheritCmd != (Tcl_Command)NULL) {
    Tcl_DeleteCommandFromToken (interp, gmlTcl_gObj_InheritCmd);
    gmlTcl_gObj_InheritCmd = (Tcl_Command)NULL;
  }

  if (gmlTcl_gObj_ThisCmd != (Tcl_Command)NULL) {
    Tcl_DeleteCommandFromToken (interp, gmlTcl_gObj_ThisCmd);
    gmlTcl_gObj_ThisCmd = (Tcl_Command)NULL;
  }

  if (gmlTcl_gObj_MethodCmd != (Tcl_Command)NULL) {
    Tcl_DeleteCommandFromToken (interp, gmlTcl_gObj_MethodCmd);
    gmlTcl_gObj_MethodCmd = (Tcl_Command)NULL;
  }

  if (gmlTcl_gObj_gmlObjectCmd != (Tcl_Command)NULL) {
    Tcl_DeleteCommandFromToken (interp, gmlTcl_gObj_gmlObjectCmd);
    gmlTcl_gObj_gmlObjectCmd = (Tcl_Command)NULL;
  }

  // Free all objects.

  while ((entry = Tcl_FirstHashEntry (&gmlTcl_gObj_ObjectTable, &search)) != (Tcl_HashEntry*)NULL) {
    theObject = (gmlTcl_TObject*)Tcl_GetHashValue (entry);
    theObject->Dispose ();
    delete theObject;
  }
  Tcl_DeleteHashTable (&gmlTcl_gObj_ObjectTable);

  // Free all classes.

  while ((entry = Tcl_FirstHashEntry (&gmlTcl_gObj_ClassTable, &search)) != (Tcl_HashEntry*)NULL) {
    theClass = (gmlTcl_TClass*)Tcl_GetHashValue (entry);
    theClass->Dispose ();
    delete theClass;
  }
  Tcl_DeleteHashTable (&gmlTcl_gObj_ClassTable);

  gmlTcl_gObj_Interp = (Tcl_Interp*)NULL;
}
