// gmlTcl_gmlObject.cpp --
//
//    Implement the "gmlObject" tcl command.
//
//  Copyright (c) 2007 LIG/IIHM
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on June 28, 2007 (FB).


#include "gmlTcl_gmlObject.h"

#include "gml_Alloc.h"

#include "gmlTcl_Obj_Init.h"
#include "gmlTcl_Obj_Internal.h"

#include "gmlTcl_Object.h"
#include "gmlTcl_Method.h"
#include "gmlTcl_Class.h"



// gmlTclObjDeleteCmd --
//
//  Implements the "gmlObject delete" tcl command.

static
int gmlTclObjDeleteCmd (Tcl_Interp*   interp,
                        int           objc,
                        Tcl_Obj*      objv[])
{
  int               res           = TCL_OK;
  int               index;
  gmlTcl_TClass*    theClass;
  gmlTcl_TObject*   theObject;
  gmlTcl_TMethod*   theMethod;

  static char *subCmds[] = {
    "class", "method", "object",
    (char *) NULL
  };
  enum ISubCmdIdx {
    IClassIdx, IMethodIdx, IObjectIdx
  };

  if (objc < 3) {
    Tcl_WrongNumArgs (interp, 2, objv, "(class|method|object) ?arg arg ...?");
    return TCL_ERROR;
  }

  res = Tcl_GetIndexFromObj (interp, objv[2], subCmds, "entity", TCL_EXACT, &index);
  if (res != TCL_OK)
    return res;

  switch (index) {
  
    case IClassIdx:
      if (objc != 4) {
        Tcl_WrongNumArgs (interp, 3, objv, "<class_name>");
        return TCL_ERROR;
      }
      theClass = gmlTcl_Obj_ClassFromTclObj (objv[3], interp);
      if (theClass == (gmlTcl_TClass*)NULL)
        return TCL_ERROR;

      theClass->Dispose ();
      delete theClass;
      break;

    case IMethodIdx:
      if (objc != 5) {
        Tcl_WrongNumArgs (interp, 3, objv, "<class_name> <method_name>");
        return TCL_ERROR;
      }
      theClass = gmlTcl_Obj_ClassFromTclObj (objv[3], interp);
      if (theClass == (gmlTcl_TClass*)NULL)
        return TCL_ERROR;

      theMethod = theClass->FindMethod (objv[4], interp, 1);
      if (theMethod == (gmlTcl_TMethod*)NULL)
        return TCL_ERROR;

      theMethod->Dispose ();
      delete theMethod;
      break;

    case IObjectIdx:
      if (objc != 4) {
        Tcl_WrongNumArgs (interp, 3, objv, "<object_name>");
        return TCL_ERROR;
      }
      theObject = gmlTcl_Obj_ObjectFromTclObj (objv[3], interp);
      if (theObject == (gmlTcl_TObject*)NULL)
        return TCL_ERROR;
      theObject->Dispose ();
      delete theObject;
      break;
  }

  return res;
}



// gmlTclObjInfoCmd --
//
//  Implements the "gmlObject info" tcl command.

static
int gmlTclObjInfoCmd (Tcl_Interp*   interp,
                      int           objc,
                      Tcl_Obj*      objv[])
{
  int                 res           = TCL_OK;
  gml_TError          err           = gml_cNoError;
  int                 index;
  gmlTcl_TEntity*     theEntity;
  gmlTcl_TClass*      theClass;
  gmlTcl_TMethod*     theMethod;
  gmlTcl_TObject*     theObject;
  Tcl_HashSearch      search;
  Tcl_HashEntry*      entry         = (Tcl_HashEntry*)NULL;
  Tcl_Obj*            resObj;
  Tcl_Obj*            resObjArray[1];
  gml_TListElem*      elem;
  Tcl_DString         ds;
  int                 superPrefix;
  gml_TList*          aList         = (gml_TList*)NULL;

  static char *subSubCmds[] = {
    "class", "object",
    (char *) NULL
  };
  enum ISubSubCmdIdx {
    IClassEntityIdx, IObjectEntityIdx
  };

  static char *subCmds[] = {
    "arglist", "args", "body",
    "class", "classes", "classofobject",
    "exists", "interface", "methods",
    "objects", "specializations",
    (char *) NULL
  };
  enum ISubCmdIdx {
    IArgListIdx, IArgsIdx, IBodyIdx,
    IClassIdx, IClassesIdx, IClassOfObjectIdx,
    IExistsIdx, IInterfaceIdx, IMethodsIdx,
    IObjectsIdx, ISpecializationsIdx
  };

  // Check arguments.

  if (objc < 3) {
    Tcl_WrongNumArgs (interp, 2, objv, "<info_type> ?arg arg ...?");
    return TCL_ERROR;
  }

  res = Tcl_GetIndexFromObj (interp, objv[2], subCmds, "info_type", TCL_EXACT, &index);
  if (res != TCL_OK)
    return res;

  switch (index) {
  
    case IArgListIdx:
    case IArgsIdx:
    case IBodyIdx:
      if (objc != 5) {
        Tcl_WrongNumArgs (interp, 3, objv, "<class_name> <method_name>");
        return TCL_ERROR;
      }
      theClass = gmlTcl_Obj_ClassFromTclObj (objv[3], interp);
      if (theClass == (gmlTcl_TClass*)NULL)
        return TCL_ERROR;

      theMethod = theClass->FindMethod (objv[4], interp, 1);
      if (theMethod == (gmlTcl_TMethod*)NULL)
        return TCL_ERROR;

      switch (index) {
        case IArgListIdx:
          Tcl_SetObjResult (interp, theMethod->ArgList ());
          break;
        case IArgsIdx:
          Tcl_SetObjResult (interp, theMethod->Args ());
          break;
        case IBodyIdx:
          Tcl_SetObjResult (interp, theMethod->Body ());
          break;
      }
      break;

    case IClassIdx:
    case IInterfaceIdx:
      if (objc != 4) {
        Tcl_WrongNumArgs (interp, 3, objv, "(<class_name>|<object_name>)");
        return TCL_ERROR;
      }
      theObject = gmlTcl_Obj_ObjectFromTclObj (objv[3]);
      if (theObject != (gmlTcl_TObject*)NULL)
        theClass = theObject->Class ();
      else {
        theClass = gmlTcl_Obj_ClassFromTclObj (objv[3]);
        if (theClass == (gmlTcl_TClass*)NULL) {
         Tcl_AppendResult (interp, "there is no class or object names \"",
                                    Tcl_GetString (objv[3]), "\"", (char*)NULL);
          return TCL_ERROR;
        }
      }

      Tcl_DStringInit (&ds);

      entry = Tcl_FirstHashEntry (theClass->MethodTable (), &search);
      while (entry != (Tcl_HashEntry*)NULL) {
        theMethod = (gmlTcl_TMethod*)Tcl_GetHashValue (entry);

      	Tcl_DStringAppend (&ds, "method ", -1);
      	Tcl_DStringAppend (&ds, theClass->ProtectedString (), -1);
      	Tcl_DStringAppend (&ds, " ", -1);
      	Tcl_DStringAppend (&ds, theMethod->ProtectedString (), -1);
      	Tcl_DStringAppend (&ds, " {", -1);
      	Tcl_DStringAppend (&ds, Tcl_GetString (theMethod->ArgList ()), -1);
        if (index == IClassIdx) {
          Tcl_DStringAppend (&ds, "} {", -1);
          Tcl_DStringAppend (&ds, Tcl_GetString (theMethod->Body ()), -1);
        }
        Tcl_DStringAppend (&ds, "}\n", -1);

        entry = Tcl_NextHashEntry (&search);
      }

      Tcl_DStringResult (interp, &ds);
      Tcl_DStringFree (&ds);
      break;

    case IClassesIdx:
      if (objc == 3) {

        // Return the list of all classes defined in this interpreter.

        res = gmlTcl_Obj_StringHashTableToObj (interp, &gmlTcl_gObj_ClassTable, &resObj, 1);
        if (res == TCL_OK)
          Tcl_SetObjResult (interp, resObj);

        return res;
      }

      // Return the list of superclass of a class or of the class of an object.

      if (objc != 4) {
        Tcl_WrongNumArgs (interp, 3, objv, "?<object_name>|<class_name>?");
        return TCL_ERROR;
      }
      theObject = gmlTcl_Obj_ObjectFromTclObj (objv[3]);
      if (theObject != (gmlTcl_TObject*)NULL)
        theClass = theObject->Class ();
      else {
        theClass = gmlTcl_Obj_ClassFromTclObj (objv[3]);
        if (theClass == (gmlTcl_TClass*)NULL) {
          Tcl_AppendResult (interp, "there is no class or object names \"",
                                    Tcl_GetString (objv[3]), "\"", (char*)NULL);
          return TCL_ERROR;
        }
      }

      err = theClass->CreateRecursiveAncestorList (&aList);
      if (err != gml_cNoError) {
        Tcl_AppendResult (interp, "could not create ancestors list: ",
                                  gml_ErrorGetMessage (err), (char*)NULL);
        return TCL_ERROR;
      }

      res = gmlTcl_Obj_EntityListToStringNameObjList (interp, aList, &resObj, 0);
      gml_TestDisposeAndDelete (aList);
      if (res != TCL_OK)
        return res;

      // Add the name of the object class if the request was on an object.

      if (theObject != (gmlTcl_TObject*)NULL) {
        resObjArray[0] = Tcl_NewStringObj (theObject->Class ()->ProtectedString (), -1);
        res = Tcl_ListObjReplace (interp, resObj, 0, 0, 1, resObjArray);
        if (res != TCL_OK) {
          Tcl_DecrRefCount (resObj);
          return res;
        }
      }
      Tcl_SetObjResult (interp, resObj);
      break;

    case IClassOfObjectIdx:
      if (objc != 4) {
        Tcl_WrongNumArgs (interp, 3, objv, "<object_name>");
        return TCL_ERROR;
      }
      theObject = gmlTcl_Obj_ObjectFromTclObj (objv[3], interp);
      if (theObject == (gmlTcl_TObject*)NULL)
        return TCL_ERROR;

      Tcl_SetObjResult (interp, Tcl_NewStringObj (theObject->Class ()->ProtectedString (), -1));
      break;
    
    case IExistsIdx:
      if (objc != 5) {
        Tcl_WrongNumArgs (interp, 3, objv, "(class|object) <name>");
        return TCL_ERROR;
      }
      res = Tcl_GetIndexFromObj (interp, objv[3], subSubCmds, "class_or_object", TCL_EXACT, &index);
      if (res != TCL_OK)
        return res;

      theEntity = gmlTcl_Obj_EntityFromTclObj ((index == IClassEntityIdx) ? gmlTcl_kClassEntityType
                                                                          : gmlTcl_kObjectEntityType, objv[4]);
      Tcl_SetObjResult (interp, Tcl_NewBooleanObj (theEntity != (gmlTcl_TEntity*)NULL));
      break;

    case IMethodsIdx:
      if ((objc < 4) || (objc > 5)) {
        Tcl_WrongNumArgs (interp, 3, objv, "(<class_name>|<object_name>) ?superclass_prefix?");
        return TCL_ERROR;
      }

      // Find the class, either from an object name or directly from the class name.

      theObject = gmlTcl_Obj_ObjectFromTclObj (objv[3]);
      if (theObject == (gmlTcl_TObject*)NULL) {
        theClass = gmlTcl_Obj_ClassFromTclObj (objv[3]);
        if (theClass == (gmlTcl_TClass*)NULL) {
          Tcl_AppendResult (interp, "there is no class or object named \"", Tcl_GetString (objv[3]),
                                    "\"", (char*)NULL);
          return TCL_ERROR;
        }
      } else
        theClass = theObject->Class ();

      // Parse the "superclass prefix" flag if provided.

      if (objc == 5) {
        res = Tcl_GetBooleanFromObj (interp, objv[4], &superPrefix);
        if (res != TCL_OK)
          return res;
      } else
        superPrefix = 0;

      // Build the list of methods, convert to a Tcl_Obj, set it as interp's result, free the list of methods.

      err = theClass->FullMethodList (&aList, superPrefix);
      if (err != gml_cNoError) {
        Tcl_AppendResult (interp, "could not create method list: ", gml_ErrorGetMessage (err), (char*)NULL);
        return TCL_ERROR;
      }

      res = gmlTcl_Obj_ListToObj (interp, aList, &resObj);
      if (res == TCL_OK)
        Tcl_SetObjResult (interp, resObj);

      while ((elem = aList->Head ()) != (gml_TListElem*)NULL) {
        free ((void*)(elem->Value ()));
        gml_DisposeAndDelete (elem);
      }
      gml_DisposeAndDelete (aList);
      break;
    
    case IObjectsIdx:
      if (objc > 4) {
        Tcl_WrongNumArgs (interp, 3, objv, "?<class_name>?");
        return TCL_ERROR;
      }

      if (objc == 3) {
        // Request for the list of all objects.

        res = gmlTcl_Obj_StringHashTableToObj (interp, &gmlTcl_gObj_ObjectTable, &resObj, 1);
        if (res == TCL_OK)
          Tcl_SetObjResult (interp, resObj);
        return res;
      }

      // Check that argument is a class name, build the list of object names for this class, set
      //  it as the interpreter's result.

      theClass = gmlTcl_Obj_ClassFromTclObj (objv[3], interp);
      if (theClass == (gmlTcl_TClass*)NULL)
        return TCL_ERROR;

      res = gmlTcl_Obj_EntityListToStringNameObjList (interp, theClass->ObjectList (), &resObj, 1);
      if (res == TCL_OK)
        Tcl_SetObjResult (interp, resObj);
      break;
    
    case ISpecializationsIdx:
      if (objc != 4) {
        Tcl_WrongNumArgs (interp, 3, objv, "<class_name>");
        return TCL_ERROR;
      }
      theClass = gmlTcl_Obj_ClassFromTclObj (objv[3], interp);
      if (theClass == (gmlTcl_TClass*)NULL)
        return TCL_ERROR;

      res = gmlTcl_Obj_EntityListToStringNameObjList (interp, theClass->Specializations (), &resObj, 0);
      if (res == TCL_OK)
        Tcl_SetObjResult (interp, resObj);
      break;
  }

  return res;
}


// gml_gmlObjectObjCmd --
//
//  Called by the interpreter as the handler of the "gmlObject" command in a script.


extern "C"
int gml_gmlObjectObjCmd  (ClientData    /* dummy */,
                          Tcl_Interp*   interp,
                          int           objc,
                          Tcl_Obj*      objv[])
{
  int               res           = TCL_OK;
  int               index;

  static char *subCmds[] = {
    "delete", "info", "unload",
    (char *) NULL
  };
  enum ISubCmdIdx {
    IDeleteIdx, IInfoIdx, IUnloadIdx
  };

  // Check arguments.

  if (objc < 2) {
    Tcl_WrongNumArgs (interp, 1, objv, "option ?arg arg ...?");
    return TCL_ERROR;
  }

  res = Tcl_GetIndexFromObj (interp, objv[1], subCmds, "option", TCL_EXACT, &index);
  if (res != TCL_OK)
    return res;

  switch (index) {
  
    case IDeleteIdx:
      res = gmlTclObjDeleteCmd (interp, objc, objv);
      break;

    case IInfoIdx:
      res = gmlTclObjInfoCmd (interp, objc, objv);
      break;

    case IUnloadIdx:
      gmlTcl_Obj_Dispose (interp);
      break;
  }

  return res;
}
